#include <gtk/gtk.h>
#include "gui.h"
#include "soundNode.h"
#include "loopStore.h"
#include "fileops.h"
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

static void save_ok_cb (GtkWidget * w, gpointer data);
static void save_cancel_cb (GtkWidget * w, gpointer data);
static void load_ok_cb (GtkWidget * w, gpointer data);
static void load_cancel_cb (GtkWidget * w, gpointer data);
static void save_sbank_ok_cb (GtkWidget * w, gpointer data);
static void load_sbank_ok_cb (GtkWidget * w, gpointer data);


static void real_save (gchar * fileName);
static void real_load (gchar * fileName);

static void real_savesbank (gchar * fileName);
static void real_loadsbank (gchar * fileName);


static inline int endian (int a) {return a;} // fix for portability later!

void SaveLayout ()
{
	// first, bring up a dialog box for save filename
	GtkWidget * fileBox = gtk_file_selection_new("Save Layout");
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->ok_button),
			"clicked",
			(GtkSignalFunc)save_ok_cb, fileBox);
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->cancel_button),
			"clicked",
			(GtkSignalFunc)save_cancel_cb, fileBox);
	gtk_widget_show(fileBox);
	
}

void LoadLayout ()
{
	// bring up a dialog box for load filename
	GtkWidget * fileBox = gtk_file_selection_new("Load Layout");
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->ok_button),
			"clicked",
			(GtkSignalFunc)load_ok_cb, fileBox);
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->cancel_button),
			"clicked",
			(GtkSignalFunc)load_cancel_cb, fileBox);
	gtk_widget_show(fileBox);
}

void SaveSBank()
{
	GtkWidget * fileBox = gtk_file_selection_new("Save Sample Bank");
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->ok_button),
			"clicked",
			(GtkSignalFunc)save_sbank_ok_cb, fileBox);
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->cancel_button),
			"clicked",
			(GtkSignalFunc)save_cancel_cb, fileBox);
	gtk_widget_show(fileBox);

}

void LoadSBank()
{
	GtkWidget * fileBox = gtk_file_selection_new("Load Sample Bank");
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->ok_button),
			"clicked",
			(GtkSignalFunc)load_sbank_ok_cb, fileBox);
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->cancel_button),
			"clicked",
			(GtkSignalFunc)load_cancel_cb, fileBox);
	gtk_widget_show(fileBox);

}


static void load_ok_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gchar * fileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileBox));
	real_load(fileName);
	gtk_widget_destroy(fileBox);
}

static void load_cancel_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gtk_widget_destroy(fileBox);
}


static void save_ok_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gchar * fileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileBox));
	real_save(fileName);
	gtk_widget_destroy(fileBox);
}

static void save_cancel_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gtk_widget_destroy(fileBox);
}

static void save_sbank_ok_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gchar * fileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileBox));
	real_savesbank(fileName);
	gtk_widget_destroy(fileBox);
}


static void load_sbank_ok_cb (GtkWidget * w, gpointer data)
{
	GtkWidget * fileBox = (GtkWidget *) data;
	gchar * fileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileBox));
	real_loadsbank(fileName);
	gtk_widget_destroy(fileBox);
}

extern mainGui * Parent;
static void real_savesbank (gchar * fileName)
{
	char ident[] = "ELE_SBANK\n";
	gchar * fName;
	gchar * p = fileName; while (*p != 0) p++;
	if (p== fileName || *(p-1) == '/' )
	{
		fName = g_strdup_printf("%sdefault.sbk", fileName);
	}
	else
	{
		fName = g_strdup(fileName); g_strreverse(fName);
		if (g_strncasecmp(fName, "kbs.", 4) != 0)
		{
			g_free(fName);
			fName = g_strdup_printf("%s.sbk", fileName);
		}
		else
		{
			g_free(fName);
			fName = g_strdup(fileName);
		}
	}
	FILE * outFile = fopen(fName, "wb");
	int i, nLoops;
	if (outFile == NULL) {cout << "Error opening file " << fName << endl; return;}
	/// write identifier
	fwrite(ident, 1, strlen(ident), outFile);
	/// write number of valid loops
	nLoops = 0;
	for (i=0;i<NUM_LOOPS;i++)
		if (isLoopFull(i)) nLoops++;
	fwrite(&nLoops, 1, 4, outFile);
	/// for each valid loop
	gchar * loopName;
	for (i=0;i<NUM_LOOPS;i++)
	{
		if (!isLoopFull(i)) continue;
		/// write loop number and name
		fwrite(&i, 1, 4, outFile);
		gtk_clist_get_text( GTK_CLIST(Parent->sampCList), i, 1, &loopName);
		fwrite(loopName, 1, strlen(loopName)+1, outFile);
		/// write loop length in bytes
		fwrite(&loopLength[i], 1, 4, outFile);
		/// write loop data
		fwrite(loopData[i], 1, loopLength[i], outFile);
	}
	
	fclose(outFile);
	g_free(fName);
}


static void real_save (gchar * fileName)
{
	char ident[] = "ELE_LAYOUT\n";
	gchar * fName;
	gchar * p = fileName; while (*p != 0) p++;
	if (p == fileName || *(p-1) == '/' )
	{
		fName = g_strdup_printf("%sdefault.lyo", fileName);
	}
	else
	{
		fName = g_strdup(fileName); g_strreverse(fName);
		if (g_strncasecmp(fName, "oyl.", 4) != 0)
		{
			g_free(fName);
			fName = g_strdup_printf("%s.lyo", fileName);
		}
		else
		{
			g_free(fName);
			fName = g_strdup(fileName);
		}
	}
	
	FILE * outFile = fopen(fName, "wb");
	if (outFile == NULL) {cout << "Error opening file " << fName << endl; return;}
	/// time to write some identifier
	fwrite(ident, 1, strlen(ident), outFile);
	/// write number of modules
	fwrite(&MAX_MODULES, 1, 4, outFile);  
	/// for each module, write stuff
	int i, j; soundModule * M;
	for (i=0;i<MAX_MODULES;i++)
	{
		M = (soundModule *) MOD[i];
		if (M == NULL)
		{
			j = -1;
			fwrite(&j, 1, 4, outFile);
		}
		else
		{
		fwrite(&M->Type, 1, 4, outFile);
		for (j=0;j<16;j++) fwrite( &M->parms[j], 1, sizeof(float), outFile);
		}
	}
	/// write connection tables
	for (i=0;i<TERMS_I_OR_O;i++)
	{
		j = (int) I_TERM_CONNECT[i]; fwrite(&j, 1, 4, outFile);
		j = (int) O_TERM_CONNECT[i]; fwrite(&j, 1, 4, outFile);
	}
	fclose(outFile);
	
	g_free(fName);
}

static void real_loadsbank (gchar * fileName)
{
	char ident[] = "ELE_SBANK\n";
	char chkId[20]; int res;
	int iTmp; 
	FILE * inFile = fopen(fileName, "rb");
	if (inFile == NULL) {cout << "Error opening file " << fileName << endl; return;}
	res = fread(chkId, 1, strlen(ident), inFile);
	if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
	if (g_strncasecmp(ident, chkId, strlen(ident) ) != 0)
	{
		cout << "Not a valid ELE Samplebank file!" << endl; fclose(inFile); return;
	}
	res = fread(&iTmp, 1, 4, inFile);
	if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
	int nLoops = iTmp; int nSlot; int lLen; char * lData;
	gchar lName[1024]; int i;
	while (nLoops > 0)
	{
		nLoops--;
		res = fread(&iTmp, 1, 4, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		nSlot = iTmp;
		g_assert(nSlot >= 0 && nSlot < NUM_LOOPS);
		i=0;
anotherchar:	res = fread(&lName[i],1,1,inFile); i++;
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		if (lName[i-1] != 0) goto anotherchar;
		gtk_clist_set_text( GTK_CLIST(Parent->sampCList), nSlot, 1, lName);
		res = fread(&lLen, 1, 4, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		lData = new char [lLen];
		res = fread(lData, 1, lLen, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		lockLoop(nSlot);
		putrawLoop(nSlot, (short *) lData, lLen/2 );
		unlockLoop(nSlot);
		
				
	}
	

}


static void real_load (gchar * fileName)
{
	char ident[] = "ELE_LAYOUT\n";
	char chkId[20]; int res;
	int iTmp; float fTmp; int i, j; 
	soundModule * newMods[64];
	int NEW_I_TERM_CONNECT[TERMS_I_OR_O];
	int NEW_O_TERM_CONNECT[TERMS_I_OR_O];
	FILE * inFile = fopen(fileName, "rb");
	if (inFile == NULL) {cout << "Error opening file " << fileName << endl; return;}
	res = fread(chkId, 1, strlen(ident) , inFile);
	if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
	if (g_strncasecmp(ident, chkId, strlen(ident) ) != 0)
	{
		cout << "Not a valid ELE Layout file!"<< endl; fclose(inFile); return;
	}
	res = fread(&iTmp, 1, 4, inFile);
	if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
	if (iTmp > MAX_MODULES)
	{cout << "File is corrupted or wrong version." << endl; fclose(inFile); return;}
	cout << "Loading module parameters" << endl;
	for (i=0;i<MAX_MODULES;i++) 
	{
		// read type of module
		res = fread(&iTmp, 1, 4, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		if (iTmp == -1)	newMods[i] = NULL;
		else
		{
			newMods[i] = new soundModule;
			newMods[i]->Type = iTmp;
			for (j=0;j<16;j++)
			{
				fread(&fTmp, 1, 4, inFile);
				newMods[i]->parms[j] = fTmp;
			}
		}
		
	}
	cout << "Loading connections" << endl;
	for (i=0;i<TERMS_I_OR_O;i++)
	{
		res = fread(&iTmp, 1, 4, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		NEW_I_TERM_CONNECT[i]=iTmp;
		res = fread(&iTmp, 1, 4, inFile);
		if (res <= 0) {cout << "Error reading file " << fileName << endl; fclose(inFile); return;}
		NEW_O_TERM_CONNECT[i]=iTmp;
	}
	fclose(inFile);
	cout << "Closing all parameter dialog boxes" << endl;
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD_PARMBOX[i]) gtk_widget_destroy(MOD_PARMBOX_HANDLE[i]);
	}
	cout << "Swapping modules and maps" << endl;
	GET_MODGRAPH_LOCK();
	for (i=0;i<TERMS_I_OR_O;i++)
	{
		I_TERM_CONNECT[i] = NEW_I_TERM_CONNECT[i];
		O_TERM_CONNECT[i] = NEW_O_TERM_CONNECT[i];
	}
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD[i] != NULL) Clean_Module(i);
		MOD[i] = newMods[i];
		if (MOD[i] != NULL) Init_Module(i);
	}
	validateModuleConnections(); // delete all non-transitive wires for safety
	assignModuleTiers();
	RELEASE_MODGRAPH_LOCK();
	forcePaint();
}


