#include <gtk/gtk.h>
#include "gui.h"
#include "soundNode.h"
#include "loopStore.h"
#include "cmdparse.h"
#include "fileops.h"
#include <iostream.h>
#include <string.h>
#include <stdlib.h>

mainGui * Parent = NULL;
void forcePaint()
{
	if (Parent != NULL)
	{
		gtk_widget_hide(Parent->drawArea);
		gtk_widget_show(Parent->drawArea);
				
	}
}


#define FORECOLOR(CC) (gdk_gc_set_foreground(widget->style->fg_gc[widget->state], &CC))
#define ALLOCCOLOR(CC) (gdk_colormap_alloc_color(cMap, &CC, FALSE, TRUE))

void parmScroll(char * D, int M);

void parmIntScroll(char * D, int M);

void parmRadio(char * D, int M);

void parmDialog(int M)
{
	gchar * Desc = typeList[MOD[M]->Type].parmDesc;
	if (Desc == NULL) return; // sanity check!
	if (Desc[0] == 0) return; // another sanity check.
	if (MOD_PARMBOX[M]) //can't have multiple boxes up for same thing!
	{		    //but the user obviously wants it up
		// this is ugly, but I couldn't find a better way to do it
		gtk_widget_hide(MOD_PARMBOX_HANDLE[M]);
		gtk_widget_show(MOD_PARMBOX_HANDLE[M]);
		return;
	}
	
	switch (Desc[0])
	{
	case '!':		//! means scrollbars controlling parms
		parmScroll(Desc+1, M);
		break;
	case '#':		//# means scrollbars which only give integer values
		parmIntScroll(Desc+1, M);
		break;
	case '%':
		parmRadio(Desc+1, M);
		break;
	}
	
}

static int delete_event_parmbox (GtkWidget * window, GdkEventAny * e, gpointer data)
{
	int M = GPOINTER_TO_INT(data);
	MOD_PARMBOX[M]=FALSE;
	return FALSE;
}

static void cb_move_parm (GtkAdjustment *adj, gpointer parm)
{
	float * v = (float *) parm;
	*v = adj->value;
}

static void cb_move_int_parm (GtkAdjustment *adj, gpointer parm)
{
	float * v = (float *) parm;
	adj->value = (float) int (adj->value);
	*v = adj->value;
}

static void cb_radio_clicked (GtkWidget *w, gpointer data)
{
	int D = GPOINTER_TO_INT(data);
	int M = D & 255;
	int N = D >> 8;
	soundModule * Me = (soundModule *) MOD[M];
	Me->parms[0]=float(N) + 0.01f;
	
}

void getfromRec(int, short **, int *);

static void cb_save_rec_button (GtkWidget *w, gpointer data)
{
	// save data from Record module to next empty sample slot
	int M = GPOINTER_TO_INT(data);
	int NextLoop;
	for (NextLoop = 1; NextLoop < NUM_LOOPS; NextLoop++)
	{if (!isLoopFull(NextLoop)) break;}
	if (NextLoop == NUM_LOOPS) return; // no free loops
	cout << "Next available loop: " << NextLoop << endl;
	lockLoop(NextLoop);
	GET_MODGRAPH_LOCK();
	// request duplicate of data from module
	short * Data; int Len;
	getfromRec(M, &Data, &Len);
	if (Len == 0) {RELEASE_MODGRAPH_LOCK(); unlockLoop(NextLoop); return;}
	putrawLoop(NextLoop, Data, Len);
	RELEASE_MODGRAPH_LOCK();
	unlockLoop(NextLoop);		
	// give new loop a name now
	gchar * lname = g_strdup_printf("Mod ID:%d", MOD[M]->UID);
	GtkWidget * sampCList = Parent->sampCList;
	gtk_clist_set_text(GTK_CLIST(sampCList), NextLoop, 1, lname);
	g_free(lname);
	cout << "Stored recorded data. " << endl;
}



void parmRadio (char * D, int M)
{
	// break D up into radio button labels
	gchar ** splitArray; int numLabels = 0;
	splitArray = g_strsplit(D, "/", 16);
	int i;
	for (i=0;i<16;i++)
	{
		if (splitArray[i]==NULL) break;
		numLabels++;
	}
	if (numLabels == 0) return;
	GtkWidget * parmWindow; gchar * title;
	title = g_strdup_printf("%s ID: %d", typeList[MOD[M]->Type].name, MOD[M]->UID);
	gpointer parmMod = GINT_TO_POINTER(M); // for button callbacks
	parmWindow = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(parmWindow), title);
	GtkWidget * label = gtk_label_new(title);
	g_free(title);
	gtk_signal_connect(GTK_OBJECT(parmWindow),
			"delete_event",
			GTK_SIGNAL_FUNC(delete_event_parmbox),
			parmMod);
	GtkWidget * parmBox;
	parmBox = gtk_vbox_new(FALSE,0);
	gtk_container_add(GTK_CONTAINER(parmWindow), parmBox);
	GSList * group;
	GtkWidget * button;
	soundModule * me = (soundModule *) MOD[M];
	int hitMe = int(me->parms[0]);
	
	button = gtk_radio_button_new_with_label(NULL, splitArray[0]);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
	gtk_box_pack_start(GTK_BOX(parmBox),label, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parmBox),button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(cb_radio_clicked), 
			GINT_TO_POINTER(M));
	if (hitMe == 0) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
	for (i=1;i<numLabels;i++)
	{
		button = gtk_radio_button_new_with_label(group, splitArray[i]);
		group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
		gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(cb_radio_clicked), 
				GINT_TO_POINTER(M + (i<<8) ));
		gtk_box_pack_start(GTK_BOX(parmBox),button, TRUE, TRUE, 0);
		if (hitMe == i) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),TRUE);
	}
	// FIXME:
	// this is supposed to be a generic dialog, but for now it'll be a Record box.

	GtkWidget * saveButton = gtk_button_new_with_label("Save Sample");
	gtk_signal_connect(GTK_OBJECT(saveButton), "clicked", GTK_SIGNAL_FUNC(cb_save_rec_button),
			GINT_TO_POINTER(M));
	gtk_box_pack_start(GTK_BOX(parmBox), saveButton, TRUE, TRUE, 0);
	gtk_widget_show_all(parmWindow);
	MOD_PARMBOX_HANDLE[M] = parmWindow;
	MOD_PARMBOX[M]=TRUE;
}


void parmScroll(char * D, int M)
{
	// break D up into variable names
	gchar ** splitArray; int numVars = 0;
	splitArray = g_strsplit(D, "/", 16);
	int i;
	for (i=0;i<16;i++)
	{
		if (splitArray[i]==NULL) break;
		numVars++;
	}
	if (numVars==0) return;
	GtkWidget * parmWindow;
	gchar * title;
	title = g_strdup_printf("%s ID: %d", typeList[MOD[M]->Type].name, MOD[M]->UID);
	
	gpointer parmMod = GINT_TO_POINTER(M); // for dialog callbacks
	parmWindow = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(parmWindow), title);
	g_free(title);
	gtk_signal_connect(GTK_OBJECT(parmWindow),
			"delete_event",
			GTK_SIGNAL_FUNC(delete_event_parmbox),
			parmMod);
	
	GtkWidget * parmTable;
	parmTable = gtk_table_new(2,numVars,FALSE);
	GtkObject* adjustment[16];
	GtkWidget* scrollbars[16];
	GtkWidget* parmlabels[16];
	
	
	/* create scrollbars and parmlabels */
	/* attach to table	*/
	for (i=0;i<numVars;i++)
	{
		adjustment[i]=gtk_adjustment_new( MOD[M]->parms[i],
			0.0f,
			256.0f,
			1.0f,
			8.0f,
			0.0f);
		gtk_signal_connect (GTK_OBJECT (adjustment[i]), 
			"value_changed",
                         GTK_SIGNAL_FUNC (cb_move_parm), 
			 (float *)&MOD[M]->parms[i]);
		scrollbars[i]=gtk_vscale_new(GTK_ADJUSTMENT(adjustment[i]));
		gtk_table_attach(GTK_TABLE(parmTable),
			scrollbars[i],
			i, i+1,
			0, 1,
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			0, 0);
			
		parmlabels[i]=gtk_label_new(splitArray[i]);
		gtk_table_attach(GTK_TABLE(parmTable),
			parmlabels[i],
			i, i+1,
			1, 2,
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			10, 0);
		g_free(splitArray[i]);
	}
	
	gtk_container_add(GTK_CONTAINER(parmWindow), parmTable);
	
	gtk_widget_show_all(parmWindow);
	MOD_PARMBOX_HANDLE[M] = parmWindow;
	MOD_PARMBOX[M]=TRUE;
}


void parmIntScroll(char *D, int M)
{
	// break D up into variable names and integer ranges
	gchar ** splitArray; int numVars = 0;
	float Start[16]; float End[16];
	splitArray = g_strsplit(D, "/", 48);
	int i;
	for (i=0;i<48;i++)
	{
		if (splitArray[i]==NULL) break;
		numVars++;
	}
	if (numVars == 0) return;
	if (numVars % 3 != 0) return;
	numVars /= 3;
	for (i=0;i<numVars;i++)
	{
		Start[i] = (float) atoi (splitArray[i*3 + 1]);
		End[i] = (float) atoi (splitArray[i*3 + 2]);
	}
	GtkWidget * parmWindow;
	gchar * title;
	title = g_strdup_printf("%s ID: %d", typeList[MOD[M]->Type].name, MOD[M]->UID);
	
	gpointer parmMod = GINT_TO_POINTER(M); // for dialog callbacks
	parmWindow = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(parmWindow), title);
	g_free(title);
	gtk_signal_connect(GTK_OBJECT(parmWindow),
			"delete_event",
			GTK_SIGNAL_FUNC(delete_event_parmbox),
			parmMod);
	GtkWidget * parmTable;
	parmTable = gtk_table_new(2,numVars,FALSE);
	GtkObject* adjustment[16];
	GtkWidget* scrollbars[16];
	GtkWidget* parmlabels[16];
	
	
	/* create scrollbars and parmlabels */
	/* attach to table	*/
	for (i=0;i<numVars;i++)
	{
		adjustment[i]=gtk_adjustment_new( MOD[M]->parms[i],
			Start[i],
			End[i],
			1.0f,
			10.0f,
			0.0f);
		gtk_signal_connect (GTK_OBJECT (adjustment[i]), 
			"value_changed",
                         GTK_SIGNAL_FUNC (cb_move_int_parm), 
			 (float *)&MOD[M]->parms[i]);
		scrollbars[i]=gtk_vscale_new(GTK_ADJUSTMENT(adjustment[i]));
		gtk_table_attach(GTK_TABLE(parmTable),
			scrollbars[i],
			i, i+1,
			0, 1,
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			0, 0);
			
		parmlabels[i]=gtk_label_new(splitArray[i*3]);
		gtk_table_attach(GTK_TABLE(parmTable),
			parmlabels[i],
			i, i+1,
			1, 2,
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			GtkAttachOptions(GTK_FILL|GTK_EXPAND|GTK_SHRINK),
			10, 0);
	}
	
	gtk_container_add(GTK_CONTAINER(parmWindow), parmTable);
	
	gtk_widget_show_all(parmWindow);
	MOD_PARMBOX_HANDLE[M] = parmWindow;
	MOD_PARMBOX[M]=TRUE;
	for (i=0;i<(numVars*3);i++)
		if (splitArray[i]!=NULL) g_free(splitArray[i]);
	 
}



// terminal selected by left-clicking
static gboolean selTerm = FALSE;
static gboolean selIsI = FALSE;	// is it I or O?
static int	selMOD = -1;	// which module?
static int	selNUM = -1;	// which terminal?

// remember last left-click event!
static gboolean oldSelTerm = FALSE;
static gboolean oldSelIsI = FALSE;
static int	oldSelMOD = -1;
static int	oldSelNUM = -1;

// handle patching of wires
static gboolean isPatchMode = FALSE;
static gboolean isPatchI = FALSE;
static int	patchMOD = -1;
static int	patchNUM = -1;

static gboolean paint_drawArea (GtkWidget *widget, GdkEventExpose *event,
				gpointer data)
{
	mainGui * Me = (mainGui *) data; 
	GdkFont * menuFont = Me->cmdLine->style->font;
	GdkColor RED, GREEN, BLUE, YELLOW, BLACK;
	GdkColormap * cMap = gtk_widget_get_colormap(widget);
	RED.red = 65535; RED.green = RED.blue = 0;
	GREEN.red = 0; GREEN.green = 65535; GREEN.blue = 0;
	BLUE.red = BLUE.green = 0; BLUE.blue = 65535;
	YELLOW.red = YELLOW.green = 65535; YELLOW.blue = 0;
	BLACK.red = BLACK.green = BLACK.blue = 0;
	g_return_val_if_fail( ALLOCCOLOR(RED), FALSE );
	g_return_val_if_fail( ALLOCCOLOR(GREEN), FALSE );
	g_return_val_if_fail( ALLOCCOLOR(BLUE), FALSE );		
	g_return_val_if_fail( ALLOCCOLOR(YELLOW), FALSE );
	g_return_val_if_fail( ALLOCCOLOR(BLACK), FALSE );

	GdkDrawable* WINDOW = widget->window;
	GdkGC * GC = widget->style->fg_gc[widget->state];
	
	gdk_window_clear_area (WINDOW,
                          event->area.x, event->area.y,
                          event->area.width, event->area.height);
	gdk_gc_set_clip_rectangle (GC, &event->area);
	
	// draw module boxes //
	int BoxX, BoxY;
	BoxX = widget->allocation.width / (MAX_TIER+1);
	BoxY = widget->allocation.height  / ((MAX_NUM_IN_TIER<1)?1:MAX_NUM_IN_TIER);
	int BXOff, BYOff;
	BXOff = int(float(BoxX)*0.1f); BYOff = int(float(BoxY)*0.1f);
	int BW, BH;
	BW = BoxX - 2*BXOff; BH = BoxY - 2*BYOff;
	int i, j, x, y; int strX, strY, bmask, tx, ty;
	gchar * mName;
	// draw boxes, labels, terminals
	
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD[i]==NULL) continue; // no module here!
		FORECOLOR(BLACK);
		x = MOD_TIER[i]; y = MOD_NUM_IN_TIER[i];
		gdk_draw_rectangle( WINDOW, GC,
			FALSE,
			BoxX*x + BXOff, BoxY*y + BYOff,
			BW, BH);
		mName = g_strdup_printf("%s ID: %d", typeList[MOD[i]->Type].name, MOD[i]->UID);
		strX = gdk_string_width(menuFont, mName);
		strY = gdk_string_height(menuFont, mName);
		// center X, Y in Box //
		gdk_draw_string( WINDOW, menuFont, GC,
				BoxX*x + (BoxX - strX)/2,
				BoxY*y + (BoxY - strY)/2,
				mName);
		g_free(mName);
		// mask is I0123 O0123
		// bmask = typeList[MOD[i]->Type].termMask;
		for (j=0;j<4;j++)
		{
			switch (j)
			{
			case 0: FORECOLOR(RED); break;
			case 1: FORECOLOR(YELLOW); break;
			case 2: FORECOLOR(GREEN); break;
			case 3: FORECOLOR(BLUE); break;
			};
			
			ty = BoxY*y + 3*BYOff/4 + (j+1)*(BH/5);
			tx = BoxX*x + 3*BXOff/4;
			bmask = 0x80 >> j;
			if (bmask & typeList[MOD[i]->Type].termMask)
			{
				if ( ( (selTerm && selIsI) && (selMOD == i) ) && (selNUM == j) )
				{
					FORECOLOR(BLACK);
					gdk_draw_rectangle(WINDOW, GC,
						FALSE,
						tx, ty,
						BXOff/2,
						BYOff/2);
				}
				else
					gdk_draw_rectangle( WINDOW, GC,
						TRUE,
						tx, ty,
						BXOff/2,
						BYOff/2);
				
			}
			
			switch (j)
			{
			case 0: FORECOLOR(RED); break;
			case 1: FORECOLOR(YELLOW); break;
			case 2: FORECOLOR(GREEN); break;
			case 3: FORECOLOR(BLUE); break;
			};
			
			
			tx += BW;
			bmask = 0x08 >> j;
			if (bmask & typeList[MOD[i]->Type].termMask)
			{
				if ( ( (selTerm && (!selIsI)) && (selMOD == i) ) && (selNUM == j) )
				{
					FORECOLOR(BLACK);
					gdk_draw_rectangle( WINDOW, GC,
						FALSE,
						tx, ty,
						BXOff/2,
						BYOff/2);
				}
				else
					gdk_draw_rectangle( WINDOW, GC,
						TRUE,
						tx, ty,
						BXOff/2,
						BYOff/2);
			}
			
			
		}
		

	}
	
	// draw wires 
	FORECOLOR(BLACK);
	int x2, y2, x1, y1;
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD[i]==NULL) continue; // no module here!
		x = MOD_TIER[i]; y = MOD_NUM_IN_TIER[i];
		// scan through output terminals
		for (j=0;j<4;j++)
		{
			if (is_O_WIRED(IO_FROM_MOD_TERM(i,j)))
			{
			x2 = MOD_TIER[MOD_FROM_IO(O_TERM_WIRE(IO_FROM_MOD_TERM(i,j)))];
			y2 = MOD_NUM_IN_TIER[MOD_FROM_IO(O_TERM_WIRE(IO_FROM_MOD_TERM(i,j)))];
			ty = O_TERM_WIRE(IO_FROM_MOD_TERM(i,j))&3;
			x1 = BoxX*x + BXOff + BW;
			y1 = BoxY*y + BYOff + (j+1)*(BH/5);
			x2 = BoxX * x2 + BXOff;
			y2 = BoxY*y2 + BYOff + ( ty + 1)*(BH/5);
			gdk_draw_line(WINDOW, GC, x1, y1, x2, y2);
			}
		}
		
	}
	
	
	gdk_gc_set_clip_rectangle (GC, NULL);
	
	FORECOLOR(BLACK);

	return TRUE;
}	



static int delete_event_cb (GtkWidget * window, GdkEventAny * e, gpointer data)
{
	gtk_main_quit();
	return FALSE;
}

static int delete_samplist_cb (GtkWidget * window, GdkEventAny * e, gpointer data)
{
	mainGui * Me = (mainGui * ) data;
	gtk_widget_hide(Me->sampWnd);
	return TRUE;
}

struct FILE_box_info
{
	GtkWidget * filebox;
	int row;
	int isOn;
};

static FILE_box_info file_box_info [NUM_LOOPS];

static int file_delete_cb (GtkWidget * window, GdkEventAny * e, gpointer data)
{
	FILE_box_info * myBox = (FILE_box_info *) data;
	file_box_info[myBox->row].isOn = 0;
	return FALSE;
}

static gchar * LastGoodFileName = NULL;
static gchar NullStr[]="";

static void file_ok_cb (GtkWidget * w, gpointer data)
{
	FILE_box_info * myBox = (FILE_box_info *) data;
	gchar * currentFile = gtk_file_selection_get_filename(GTK_FILE_SELECTION(
			myBox->filebox));
	int row = myBox->row;
	lockLoop(row);
	if (loadLoop(row, currentFile) == 0)
	{
		currentFile = "";
	}
	else
	{
		if (LastGoodFileName != NULL) g_free(LastGoodFileName);
		LastGoodFileName = g_strdup(currentFile);
	}
	
	unlockLoop(row);
	GtkWidget * sampCList = Parent->sampCList;
	gtk_clist_set_text(GTK_CLIST(sampCList), row, 1, currentFile);
	file_box_info[row].isOn = 0;
	gtk_widget_destroy(myBox->filebox);
}

static void file_cancel_cb (GtkWidget *w, gpointer data)
{
	FILE_box_info * myBox = (FILE_box_info *) data;
	file_box_info[myBox->row].isOn = 0;
	gtk_widget_destroy(myBox->filebox);
}


static void select_sample_cb (GtkWidget * widget, gint row, gint column,
		GdkEventAny * e, gpointer data)
{
	mainGui * Me = (mainGui *) data;
	if (file_box_info[row].isOn) return;
	file_box_info[row].isOn = 1;
	gchar * currentFile;
	gtk_clist_get_text ( GTK_CLIST(Me->sampCList),
			row, 1, &currentFile);
	
	GtkWidget * fileBox;
	fileBox = gtk_file_selection_new("Load sample");
	if (currentFile[0] == 0)
	{
		currentFile = (LastGoodFileName == NULL) ? NullStr : LastGoodFileName;
	}
	
	gtk_file_selection_set_filename( GTK_FILE_SELECTION(fileBox),
				currentFile);
	
	file_box_info[row].filebox = fileBox;
	file_box_info[row].row = row;
	
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->ok_button),
			"clicked",
			(GtkSignalFunc)file_ok_cb, &file_box_info[row]);
	
	gtk_signal_connect(  GTK_OBJECT( GTK_FILE_SELECTION(fileBox)->cancel_button),
			"clicked",
			(GtkSignalFunc) file_cancel_cb,
			&file_box_info[row]);
	
	gtk_signal_connect(  GTK_OBJECT(fileBox),
			"delete_event",
			(GtkSignalFunc) file_delete_cb,
			&file_box_info[row]);
	gtk_widget_show(fileBox);
			
}


static void enter_cmdLine( GtkWidget *widget,
                          GtkWidget *entry )
{
	gchar *entry_text;
	entry_text = gtk_entry_get_text(GTK_ENTRY(entry));
	gtk_entry_set_text(GTK_ENTRY(entry),"");
	cmdparse(entry_text);
	
}

static gint mouse_click_cb (GtkWidget * widget, GdkEventButton * e)
{
	gdouble mx, my;
	mx = e->x; my = e->y;
	int button = e->button;
	
	
	int BoxX, BoxY;
	BoxX = widget->allocation.width / (MAX_TIER+1);
	BoxY = widget->allocation.height  / ((MAX_NUM_IN_TIER<1)?1:MAX_NUM_IN_TIER);
	int BXOff, BYOff;
	BXOff = int(float(BoxX)*0.1f); BYOff = int(float(BoxY)*0.1f);
	int BW, BH;
	BW = BoxX - 2*BXOff; BH = BoxY - 2*BYOff;
	int XCell = int(mx/BoxX); int YCell = int(my/BoxY);
	int RemX = int(mx) - XCell*BoxX; int RemY = int(my) - YCell*BoxY;
	// use right-click to bring up dialog box
	if (button == 3)
	{
		// make sure that it's actually within the box
		if (RemX < BXOff) return TRUE;
		if (RemY < BYOff) return TRUE;
		if (RemX > BXOff+BW) return TRUE;
		if (RemY > BYOff+BH) return TRUE;
		// find module in cell, if it exists
		int M, i; M = -1;
		for (i=0;i<MAX_MODULES;i++)
		{
			if (MOD[i]==NULL) continue;
			if (MOD_TIER[i]==XCell) 
			if (MOD_NUM_IN_TIER[i] == YCell) M = i;
		}
		if (M == -1) return TRUE;
		parmDialog(M);
		return TRUE;
	}
	// use left-click to select a terminal
	int M, i, bitmask;
	if (button == 1)
	{
		oldSelTerm = selTerm; oldSelIsI = selIsI; oldSelMOD = selMOD; oldSelNUM = selNUM;
		selTerm = FALSE;
		// are we in a terminal strip?
		if (RemX < BXOff/2) goto PickEmpty;
		if (RemX > BW + BXOff + BXOff/2) goto PickEmpty;
		if (RemY < BYOff) goto PickEmpty;
		if (RemY > BYOff + BH) goto PickEmpty;
		if ( (RemX > BXOff * 2) && (RemX < BW) ) goto PickEmpty;
		
		// find module in cell, if it exists
		M = -1;
		for (i=0;i<MAX_MODULES;i++)
		{
			if (MOD[i]==NULL) continue;
			if (MOD_TIER[i]==XCell) 
			if (MOD_NUM_IN_TIER[i] == YCell) M = i;
		}
		if (M == -1) goto PickEmpty;
		selMOD = M; selIsI = (RemX < BoxX/2);
		selNUM = RemY - BYOff; selNUM -= BH/10; 
		selNUM /= (BH/5);
		if (selNUM < 0) goto PickEmpty;
		if (selNUM > 3) goto PickEmpty;
		// check bitmask
		bitmask = selIsI ? 0x80 : 0x08;
		bitmask >>= selNUM;
		if ( (bitmask & typeList[MOD[M]->Type].termMask) == 0) goto PickEmpty;
		selTerm = TRUE;
	}
PickEmpty:
	// okay, now was that a connection to be made?
	int ModFrom, ModTo, NumFrom, NumTo, tmp;
	if (button == 1)
		if ( (oldSelTerm == TRUE) && (selTerm == TRUE) )
			if (oldSelIsI == !selIsI)
	{
		ModFrom = oldSelMOD; ModTo = selMOD;
		NumFrom = oldSelNUM; NumTo = selNUM;
		if (oldSelIsI) // reverse them
		{
			tmp = ModFrom; ModFrom = ModTo; ModTo = tmp;
			tmp = NumFrom; NumFrom = NumTo; NumTo = tmp;
		}
		if (!isPatchMode)
		{	// make a wire from output ModFromNumFrom to input ModToNumTo
			gchar * command;
			if (MOD[ModFrom] == NULL) goto DoneC;
			if (MOD[ModTo] == NULL) goto DoneC;
			ModFrom = MOD[ModFrom]->UID; ModTo = MOD[ModTo]->UID;
			// compose command string and invoke parser
			// wasteful, yes, but convenient for now!
			command = g_strdup_printf("C %d-%d%c%c\n", ModFrom, ModTo,
					(char) (NumFrom+'a'),
					(char) (NumTo+'a') );
			cmdparse(command);
			g_free(command);
		}
		else
		{	// remove wire for terminal in patchMOD, NUM, isI, patch new into
			isPatchMode = FALSE;
			int old_from_NUM, old_from_MOD, old_to_NUM, old_to_MOD, tmpIO;
			if (isPatchI)
			{
				old_to_MOD = patchMOD; old_to_NUM = patchNUM;
				tmpIO = I_TERM_WIRE(IO_FROM_MOD_TERM(patchMOD, patchNUM));
				if (tmpIO == -1) // wire didn't exist after all, leave
						goto DoneC;
				old_from_MOD = MOD_FROM_IO(tmpIO); old_from_NUM = tmpIO & 3;
			}
			else
			{
				old_from_MOD = patchMOD; old_from_NUM = patchNUM;
				tmpIO = O_TERM_WIRE(IO_FROM_MOD_TERM(patchMOD, patchNUM));
				if (tmpIO == -1) // another non-existent wire to patch!
						goto DoneC;
				old_to_MOD = MOD_FROM_IO(tmpIO); old_to_NUM = tmpIO & 3;
			}
			// survived this far, now, actually do wire removal
			// and replacement
			GET_MODGRAPH_LOCK();
			if (isPatchI)	cut_OTERM(tmpIO);
			else		cut_ITERM(tmpIO);
			cut_OTERM( IO_FROM_MOD_TERM(ModFrom, NumFrom));
			cut_ITERM( IO_FROM_MOD_TERM(ModTo, NumTo) );
			make_IO_WIRE( IO_FROM_MOD_TERM(old_to_MOD, old_to_NUM),
				      IO_FROM_MOD_TERM(ModFrom, NumFrom));
			make_IO_WIRE( IO_FROM_MOD_TERM(ModTo, NumTo),
				      IO_FROM_MOD_TERM(old_from_MOD, old_from_NUM));
			// <PARANOIA>
			validateModuleConnections();
			// </PARANOIA>
			assignModuleTiers();
			RELEASE_MODGRAPH_LOCK();
			
		}
					
	DoneC:	oldSelTerm = selTerm = FALSE;
	}
			
		
		 
	
	forcePaint();
	return TRUE;
}


mainGui::mainGui()
{

	mainWnd = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	vBox = gtk_vbox_new(FALSE, 0);
	makeMenu (mainWnd, &menuBar);
	
	drawArea = gtk_drawing_area_new();
	gtk_drawing_area_size (GTK_DRAWING_AREA (drawArea),
			500, 400);
	cmdLine = gtk_entry_new();
	
	gtk_box_pack_start( GTK_BOX(vBox),
			menuBar,
			FALSE, FALSE, 0);
	gtk_box_pack_start( GTK_BOX(vBox),
			drawArea,
			TRUE, TRUE, 0);
	gtk_box_pack_start( GTK_BOX(vBox),
			cmdLine,
			FALSE, FALSE, 0);
	
	gtk_container_add(GTK_CONTAINER(mainWnd), vBox);
	gtk_window_set_title(GTK_WINDOW(mainWnd), MAINGUI_TITLE);
	gtk_signal_connect (GTK_OBJECT(drawArea),
			"expose_event",
			GTK_SIGNAL_FUNC(paint_drawArea),
			this);
	
	gtk_signal_connect (GTK_OBJECT(drawArea),
			"button_press_event",
			GTK_SIGNAL_FUNC(mouse_click_cb),
			this);
	
	gtk_widget_set_events ( drawArea, GDK_EXPOSURE_MASK|
			GDK_LEAVE_NOTIFY_MASK |
			GDK_BUTTON_PRESS_MASK |
			GDK_POINTER_MOTION_MASK);
	
	
	gtk_signal_connect(GTK_OBJECT(cmdLine), "activate",
                            GTK_SIGNAL_FUNC(enter_cmdLine),
                            cmdLine);
	
	
	gtk_signal_connect (GTK_OBJECT(mainWnd), 
			"delete_event",
			GTK_SIGNAL_FUNC(delete_event_cb),
			NULL);
		
	gtk_widget_show_all(mainWnd);
	Parent = this;	
	
	sampWnd = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title (GTK_WINDOW(sampWnd), "Sample List");
	sampScrollWnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sampScrollWnd),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gchar A[] = "Number";
	gchar B[] = "FileName";
	gchar * C[2] = {A, B};
	sampCList = gtk_clist_new_with_titles(2, C);
	gtk_clist_set_selection_mode (GTK_CLIST(sampCList), GTK_SELECTION_SINGLE);
	gtk_clist_set_shadow_type (GTK_CLIST(sampCList), GTK_SHADOW_OUT);
	gtk_clist_set_column_width(GTK_CLIST(sampCList), 0, 50);
	gtk_clist_set_column_width(GTK_CLIST(sampCList), 1, 350);
	// set up number of rows //
	int row;
	gchar NUM[256];
	gchar TXT[] = "";
	gchar * D[2] = {NUM, TXT};
	for (row = 0; row < NUM_LOOPS; row++)
	{
		g_snprintf(NUM, 256, "%d", row);
		gtk_clist_append(GTK_CLIST(sampCList), D);
		file_box_info[row].isOn = 0;
	}
	
	gtk_container_set_border_width(GTK_CONTAINER(sampScrollWnd), 10);
	gtk_container_add( GTK_CONTAINER(sampScrollWnd), sampCList);
	gtk_container_add( GTK_CONTAINER(sampWnd), sampScrollWnd);
	
	gtk_window_set_default_size ( GTK_WINDOW(sampWnd), 200, 400);
	gtk_signal_connect (GTK_OBJECT(sampWnd),
			"delete_event",
			GTK_SIGNAL_FUNC(delete_samplist_cb),
			this);
	gtk_signal_connect (GTK_OBJECT(sampCList),
			"select_row",
			GTK_SIGNAL_FUNC(select_sample_cb),
			this);
	
//	gtk_widget_show_all(sampWnd);
}

mainGui::~mainGui()
{
}

static void issue_cmd( GtkWidget *w,
                              gpointer   data )
{
	gchar * Line = g_strdup( (gchar*) data);
	cmdparse( Line);
	g_free(Line);
}

static void mnu_disconnect( GtkWidget *w, gpointer data)
{
	if (selTerm)
	{
		GET_MODGRAPH_LOCK();
		if (selIsI) cut_ITERM(IO_FROM_MOD_TERM(selMOD, selNUM));
		else cut_OTERM(IO_FROM_MOD_TERM(selMOD, selNUM));
		selTerm = oldSelTerm = FALSE;
		RELEASE_MODGRAPH_LOCK();
		forcePaint();
	}
	
}

static void mnu_remove (GtkWidget *w, gpointer data)
{
	if (!selTerm) return;
	int UID = MOD[selMOD]->UID;
	gchar * command;
	command = g_strdup_printf("R%d",UID);
	cmdparse(command);
	g_free(command);
}

static void mnu_patch (GtkWidget *w, gpointer data)
{
	if (!selTerm) return;
	selTerm = FALSE;
	isPatchMode = TRUE;
	isPatchI = selIsI; patchMOD = selMOD; patchNUM = selNUM;
	forcePaint();
}


static void mnu_samplist (GtkWidget *w, gpointer data)
{
	gtk_widget_show_all( Parent->sampWnd);
}

static void mnu_savelayout (GtkWidget *w, gpointer data)
{
	SaveLayout();
}

static void mnu_openlayout (GtkWidget *w, gpointer data)
{
	LoadLayout();
}

static void mnu_opensbank (GtkWidget *w, gpointer data)
{
	LoadSBank();
}

static void mnu_savesbank (GtkWidget *w, gpointer data)
{
	SaveSBank();
}


static GtkItemFactoryEntry mainGui_menu[] = 
{
       { "/_File",         "<alt>F",         NULL, 0, "<Branch>" },
       { "/File/_Open Layout",    "<control>O", (GtkSignalFunc) mnu_openlayout, 0, NULL },
       { "/File/_Save Layout",    "<control>S", (GtkSignalFunc) mnu_savelayout, 0, NULL },
       { "/File/Open Sa_mplebank", NULL, (GtkSignalFunc) mnu_opensbank, 0, NULL},
       { "/File/Save Sam_plebank", NULL, (GtkSignalFunc) mnu_savesbank, 0, NULL},
       { "/File/sep1",     NULL,         NULL, 0, "<Separator>" },
       { "/File/E_xit",     NULL, (GtkSignalFunc) gtk_main_quit, 0, NULL },
       { "/_Insert",      NULL,         NULL, 0, "<Branch>" },
       { "/Insert/_Passthrough",  NULL,  (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a2") , NULL },
       { "/Insert/_Mixer",  NULL,  (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a3") , NULL },
       { "/Insert/_Distortion",  NULL,  (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a4") , NULL },
       { "/Insert/_Reverb",  NULL,  (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a5") , NULL },
       { "/Insert/_Chorus", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a16"), NULL },
       { "/Insert/_Splitter",  NULL,  (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a6") , NULL },              
       { "/Insert/_Tone", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a7"), NULL},
       { "/Insert/_Looper", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a8"),NULL},
       { "/Insert/M_ultiply", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a19"), NULL},
       { "/Insert/_XFader", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a9"),NULL},
       { "/Insert/_Volume Compressor", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a17"), NULL},
       { "/Insert/_OctaveDown", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a10"),NULL},
       { "/Insert/S_kratcher", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a18"),NULL},
       { "/Insert/Record _1 Second", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a11"), NULL},
       { "/Insert/Record 1_0 Seconds", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a12"), NULL},
       { "/Insert/Record _30 Seconds", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a13"), NULL},
       { "/Insert/S_ync Sequencer", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a14"), NULL},
       { "/Insert/Pa_nning Mixer", NULL, (GtkSignalFunc)issue_cmd, GPOINTER_TO_INT("a15"), NULL},
       { "/_Disconnect",   NULL, (GtkSignalFunc) mnu_disconnect, 0, NULL},
       { "/_Remove", NULL, (GtkSignalFunc) mnu_remove, 0, NULL},
       { "/_Patch", NULL, (GtkSignalFunc) mnu_patch, 0, NULL},
       { "/_SampleList", NULL, (GtkSignalFunc) mnu_samplist, 0, NULL},
       { "/_Help",         NULL,         NULL, 0, "<LastBranch>" },
       { "/_Help/About",   NULL,         NULL, 0, NULL }
};


void mainGui::makeMenu(GtkWidget * W, GtkWidget ** MBar)
{
       GtkItemFactory *item_factory;
       GtkAccelGroup *accel_group;
       gint nmenu_items = sizeof (mainGui_menu) / sizeof (mainGui_menu[0]);

       accel_group = gtk_accel_group_new ();

       /* This function initializes the item factory.
          Param 1: The type of menu - can be GTK_TYPE_MENU_BAR, GTK_TYPE_MENU,
                   or GTK_TYPE_OPTION_MENU.
          Param 2: The path of the menu.
          Param 3: A pointer to a gtk_accel_group.  The item factory sets up
                   the accelerator table while generating menus.
       */

       item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
                                            accel_group);

       /* This function generates the menu items. Pass the item factory,
          the number of items in the array, the array itself, and any
          callback data for the the menu items. */
       gtk_item_factory_create_items (item_factory, nmenu_items, mainGui_menu, NULL);

       /* Attach the new accelerator group to the window. */
       gtk_window_add_accel_group (GTK_WINDOW (W), accel_group);

       if (MBar)
         /* Finally, return the actual menu bar created by the item factory. */ 
         *MBar = gtk_item_factory_get_widget (item_factory, "<main>");
}
