#include <gtk/gtk.h>
#include "gui.h"
#include <iostream.h>
#include <string.h>
#include "loopStore.h"
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>


char * loopData[NUM_LOOPS];
int loopLength[NUM_LOOPS];
pthread_mutex_t loopMutex[NUM_LOOPS];


void initLoops()
{
	int i;
	cout << "Creating loops and loop mutexes" << endl;
	for (i=0;i<NUM_LOOPS;i++)
	{
		loopData[i] = NULL;
		loopLength[i] = 0;
		pthread_mutex_init(&loopMutex[i],NULL);
	}
	
}

void cleanLoops()
{
	int i;
	for (i=0;i<NUM_LOOPS;i++)
	{
		pthread_mutex_destroy(&loopMutex[i]);
	}
	cout << "Destroyed loop mutexes" << endl;
}

void lockLoop(int L)
{
	pthread_mutex_lock(&loopMutex[L]);
}

void unlockLoop(int L)
{
	pthread_mutex_unlock(&loopMutex[L]);
}

gboolean isLockLoop(int L)
{
	int ret;
	ret = pthread_mutex_trylock(&loopMutex[L]);
	if (ret == 0) pthread_mutex_unlock(&loopMutex[L]);
	return (ret != 0);
}
 

void deleteLoop(int L)
{
	if (loopData[L]!=NULL)
	{
		g_free(loopData[L]);
		loopData[L]=NULL;
	}
	
}

gboolean isLoopFull (int L)
{
	return (loopData[L] != NULL);
	
}


static inline short conv8 (char c)
{
	short ret = c; ret -= 0x80; ret <<= 8; return ret;
}

static inline short av16 (short a, short b)
{
	int sum = a + b; sum /=2; return (short) sum;
}


static inline short GetNext16Mono (char * &raw, int Channels, int Bits)
{
	short a[2];
	int i = 0;
	for (i=0;i<Channels;i++)
	{
		switch (Bits)
		{
			case 8:
				a[i] = conv8( *raw);
				raw++;
				break;
			case 16:
				a[i] = * ( (short *) raw);
				raw+=2;
				break;
		};
	}
	if (Channels == 1) return a[0];
	return av16( a[0], a[1] );
	
}


int loadLoop(int L, gchar * fileName)
{
	deleteLoop(L);
	// now we have to figure out the length of the file, open it,
	// get the header to figure out format and data offset,
	// allocate proper amount of space, and convert file's WAV format
	// to 16-bit 44.1K mono
	
	// to get file length, we must open and close it first with low-level routines
	// so that we can use lseek instead of fseek (which doesn't give length)
	int fd = open(fileName, O_RDONLY);
	if (fd == -1)
	{
		cout << "Error opening file:" << fileName << endl;
		return 0;
	}
	int fileLength = lseek(fd, 0, SEEK_END);
	if (fileLength == -1)
	{
		cout << "Error determining file length:" << fileName << endl;
		return 0;
	}
	close(fd);
	
	// open the file with standard I/O library
	FILE * wavFile = fopen(fileName, "rb");
	if (wavFile == NULL)
	{
		cout << "Error opening file:" << fileName << endl;
		return 0;
	}
	char wavHdr[512]; memset (wavHdr, 0, 512);
	unsigned int hdrLen = (fileLength < 512) ? fileLength : 512;
	if (fread( wavHdr, 1, hdrLen, wavFile) != hdrLen)
	{
		cout << "Error reading header from file: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	// is it a WAVE file?
	if ( (strncmp (&wavHdr[0], "RIFF", 4) != 0) || (strncmp (&wavHdr[8], "WAVE", 4) != 0) )
	{
		cout << "Not a wave file: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	
	// now search for "fmt " tag
	int fmtOffset = -1;
	int i;
	for (i=0;i<508;i++)
	{
		if (strncmp (&wavHdr[i], "fmt ", 4) == 0)
		{
			fmtOffset = i; break;
		}
	}
	if (fmtOffset == -1)
	{
		cout << "fmt tag not found in: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	
	// read format out of file
	int Bits, HZ, Channels;
	short format = * ( (short *) &wavHdr[fmtOffset + 8] );
	if (format != 1)
	{
		cout << "Not a PCM wave file: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	Channels = * ( (short *) &wavHdr[fmtOffset + 10] );
	HZ = * ( (int *) &wavHdr[fmtOffset + 12] );
	Bits = * ( (short *) &wavHdr[fmtOffset + 22] );
	cout << "Channels " << Channels << " Bits " << Bits << " HZ " << HZ << endl;
	
	switch (HZ)
	{
		case 11025:
		case 22050:
		case 44100:
			break;
		default:
			cout << "Illegal HZ" << endl;
		fclose(wavFile);
		return 0;
	};
	
	switch (Channels)
	{
		case 1:
		case 2:
			break;
		default:
			cout << "Illegal Channels" << endl;
		fclose(wavFile);
		return 0;
	};
	
	switch (Bits)
	{
		case 8:
		case 16:
			break;
		default:
			cout << "Illegal Bits" << endl;
		fclose(wavFile);
		return 0;
	};
	
	
	int dataOffset = -1;
	// now search for "data" tag
	for (;i<508;i++)
	{
		if (strncmp (&wavHdr[i], "data", 4) == 0)
		{
			dataOffset = i; break;
		}
		
	}
	
	if (dataOffset == -1)
	{
		cout << "data tag not found in: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	
	dataOffset += 8;
	if (dataOffset >= fileLength)
	{
		cout << "File is missing data or corrupted: " << fileName << endl;
		fclose(wavFile);
		return 0;
	}
	// now read in the data in the original format
	int rawSize = fileLength - dataOffset;
	char * rawData = (char *)g_malloc(rawSize);
	fseek(wavFile, dataOffset, SEEK_SET);
	if (fread (rawData, 1, (size_t) rawSize, wavFile) != (size_t)rawSize)
	{
		cout << "Error reading data from file: " << fileName << endl;
		g_free(rawData);
		fclose(wavFile);
		return 0;
	}
	
	int loopSize = rawSize;
	loopSize /= Channels; // stereo is averaged to mono
	loopSize *= (44100 / HZ); // lower rates are padded
	loopSize *= 16/Bits; // 8-bit samples are converted to 16
	
	loopData[L] = (char *) g_malloc(loopSize);
	short * pLoop = (short *) loopData[L];
	char * raw = rawData;
	char * endraw = rawData + rawSize;
	short Get; int padHZ;
	padHZ = 44100/ HZ;
	while (raw < endraw)
	{
		Get = GetNext16Mono(raw, Channels, Bits);
		for (i=0;i<padHZ;i++)
		{
			*pLoop = Get;
			pLoop++;
		}
		
	}
	
	cout << "WAV data copied" << endl;
	loopLength[L] = loopSize;
	g_free(rawData);
	fclose(wavFile);
	return 1;
}

void putrawLoop( int L, short * raw, int nraw)
{
	deleteLoop(L);
	loopLength[L]=nraw * sizeof(short);
	loopData[L] = new char [loopLength[L]];
	memcpy ( loopData[L], (char *) raw, loopLength[L]);
	
}
