#include <gtk/gtk.h>
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "soundNode.h"
#include "gui.h"
#include "loopStore.h"
#include "cardfeed.h"
#include "backbuf.h"
#include "pump.h"



void SendOut(int M)
{
	// sends buffers out to next modules
	int OM;
	soundModule * me = (soundModule *)MOD[M];
	soundModule * other;
	int i;
	for (i=0;i<4;i++)
	{
		if (me->chunkIn[i]!=NULL)
		{
			OM = O_TERM_WIRE(IO_FROM_MOD_TERM(M,i));
			if (OM != -1)
			{
				other = (soundModule *)MOD[MOD_FROM_IO(OM)];
				other->chunkIn[OM&3]=me->chunkIn[i];
			}
			else
			{
				freeChunk(me->chunkIn[i]);
			}
			
			me->chunkIn[i]=NULL;
		}
	}
	
}

char * verifyIn(int M, int ITERM)
{
	//	if no input, then gimme some silence //
	soundModule * me = (soundModule *) MOD[M];
	if (me->chunkIn[ITERM]==NULL)
	{
		me->chunkIn[ITERM]=newChunk();
		memset(me->chunkIn[ITERM],0,sound_chunk_size);
	}
	return me->chunkIn[ITERM];
}


void fLineIn(int ME)
{
//	cout << "Line In module " << MOD[ME]->UID << endl;
	/* allocate two buffers for output */
	soundModule *me = (soundModule *)MOD[ME];
//	cout << "Give me memory" << endl;
	me->chunkIn[0]=newChunk();
	me->chunkIn[1]=newChunk();
	me->chunkIn[2]=me->chunkIn[3]=NULL;
//	cout << "Give me death" << endl;
	/* sort output from buf - alternate shorts */
	short * O1 = (short *) me->chunkIn[0];
	short * O2 = (short *) me->chunkIn[1];
	short * IN = (short *) buf;
	short * END = (short *) (buf + sound_buffer_size);
//	cout << "Do the splits" << endl;
	while (IN < END)
	{
		*O1 = *IN; IN++; O1++;
		*O2 = *IN; IN++; O2++;
		
	}
	SendOut(ME);
	/* for debugging, to make sure that buf doesn't go out unchanges */
	memset (buf, 0, sound_buffer_size);
//	cout << "Bye linein" << endl;
}

void fLineOut (int ME)
{
//	cout << "Line Out module " << MOD[ME]->UID << endl;
	/* mix the inputs */
	short * I1 = (short *) verifyIn(ME,0);
	short * I2 = (short *) verifyIn(ME,1);	
	short * O = (short *) buf;
	short * END = (short *) (buf + sound_buffer_size);
	short * KillMe1 = I1; short * KillMe2 = I2;
	while (O < END)
	{
		*O = *I1; O++; I1++;
		*O = *I2; O++; I2++;
	}
	freeChunk( (char*) KillMe1);
	freeChunk( (char*) KillMe2);
	MOD[ME]->chunkIn[0]=MOD[ME]->chunkIn[1]=NULL;
//	cout << "Bye lineout" << endl;
}

void iMixer (int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	int j;
	for (j=0;j<4;j++) me->parms[j]=200.0f;
}


void fMixer (int ME)
{
//	cout << "Mixer module " << MOD[ME]->UID << endl;
	int i;
	int nSamps = sound_chunk_size/2;
	int tmp, total;
	short * In[4]; int iVol[4];
	soundModule *me = (soundModule *)MOD[ME];
	for (i=0;i<4;i++)
	{
		In[i]=(short *)verifyIn(ME,i);
		iVol[i]=(int) me->parms[i];
		iVol[i]= (iVol[i]<0) ? 0: (iVol[i]>256) ? 256 : iVol[i];
	}
	for (i=0;i<nSamps;i++)
	{
		tmp = *In[0]; tmp*=iVol[0]; tmp /= 256; total = tmp;
		tmp = *In[1]; tmp*=iVol[1]; tmp /= 256; total += tmp;
		tmp = *In[2]; tmp*=iVol[2]; tmp /= 256; total += tmp;
		tmp = *In[3]; tmp*=iVol[3]; tmp /= 256; total += tmp;
		*In[0] = (total < -32768) ? -32768 : (total > 32767) ? 32767 : total;
		In[0]++; In[1]++; In[2]++; In[3]++;
	}
	for (i=1;i<4;i++)
	{
		freeChunk( (char*) me->chunkIn[i]); me->chunkIn[i]=NULL;
	}
	SendOut(ME);	
}

void iPass (int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	int j;
	for (j=0;j<4;j++) me->parms[j]=200.0f;
}


void fPass (int ME)
{
//	cout << "Passthrough module " << MOD[ME]->UID << endl;
	int i;
	short * In[4]; 
	soundModule *me = (soundModule *)MOD[ME];
	for (i=0;i<4;i++)
		In[i]=(short *)verifyIn(ME,i);
	int iVol[4];
	for (i=0;i<4;i++)
	{
		iVol[i]=(int) me->parms[i];
		iVol[i]= (iVol[i]<0) ? 0: (iVol[i]>256) ? 256 : iVol[i];
	}

	int nSamps = sound_chunk_size/2;
	int tmp;
	for (i=0;i<nSamps;i++)
	{
		tmp = *In[0]; tmp*=iVol[0]; tmp /= 256; *In[0]=tmp; In[0]++;
		tmp = *In[1]; tmp*=iVol[1]; tmp /= 256; *In[1]=tmp; In[1]++;
		tmp = *In[2]; tmp*=iVol[2]; tmp /= 256; *In[2]=tmp; In[2]++;
		tmp = *In[3]; tmp*=iVol[3]; tmp /= 256; *In[3]=tmp; In[3]++;
		
	}
	SendOut(ME);	
}

void iDistort (int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	me->parms[0]=50.0f;
	me->parms[2]=200.0f;
}


void fDistort (int ME)
{
	short * In;
	soundModule *me = (soundModule *)MOD[ME];
	In = (short *)verifyIn(ME,0);
	int Drive = (int) me->parms[0]; 
	int Sat = (int) me->parms[1];
	// Sat clips derivative by limiting difference between samples.
	// use Extra member to store last sample for seamlessness between chunks.
	
	Sat = 65536 * 16 / (Sat + 1);
	int Level = (int) me->parms[2];
	int nSamps = sound_chunk_size/2;
	int i, tmp; int lastVal = GPOINTER_TO_INT(me->Extra);	
	for (i=0;i<nSamps;i++)
	{
		// apply Drive 
		tmp = *In; tmp*=Drive; tmp /= 16;
		// apply Sat
		if ( (tmp - lastVal) > Sat) tmp = lastVal + Sat;
		else if ( (lastVal - tmp) > Sat) tmp = lastVal - Sat;
		lastVal = tmp;
		// apply Clip
		tmp = (tmp > 32767) ? 32767 : (tmp < -32767) ? -32767 : tmp;
		// apply Level
		tmp *= Level; tmp /= 256;
		*In = tmp; In++;
	}
	
	me->Extra = GINT_TO_POINTER(lastVal);
	SendOut(ME);
}

int reverbBuffer::numChunks() { return nChunks;}

reverbBuffer::reverbBuffer(int chunks)
{
	data = (char *) g_malloc0 (chunks * sound_chunk_size);
	nChunks = chunks;
	nCursor = 0;
}

reverbBuffer::~reverbBuffer()
{
	g_free(data); data = NULL;
}

void reverbBuffer::addChunk(char * chunk)
{
	char * addTo = data + nCursor*sound_chunk_size;
	memcpy (addTo, chunk, sound_chunk_size);
	nCursor++; nCursor %= nChunks;
}

char* reverbBuffer::getChunk(int delay)
{
	g_assert( (delay >= 0) && (delay < nChunks) );
	int nFrom = nCursor - delay;
	while (nFrom < 0) nFrom += nChunks;
	char * getFrom = data + nFrom*sound_chunk_size;
	char * giveTo = newChunk();
	memcpy(giveTo, getFrom, sound_chunk_size);
	return giveTo;
}


void iReverb (int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	me->parms[0]=15.0f;
	me->parms[1]=128.0f;
	me->parms[2]=200.0f;
	me->parms[3]=0.0f;
	reverbBuffer * my_buffer;
	int NUM_CHUNKS = ( 44100 * sizeof(short) / sound_chunk_size ) / 4; //250 ms
	my_buffer = new reverbBuffer(NUM_CHUNKS + 1);
	me->Extra = (void *) my_buffer;
}


void fReverb (int ME)
{
	// Reverb up to 250 milliseconds
	soundModule *me = (soundModule *)MOD[ME];
	reverbBuffer * my_buffer;
	g_assert(me->Extra != NULL);
	my_buffer = (reverbBuffer *) me->Extra;
	
	// get delay
	float delay = me->parms[0]; 
	delay /= 256.0; delay *= (float) my_buffer->numChunks();
	int dd = (int) delay; 
	dd = (dd < 1) ? 1 : (dd >= my_buffer->numChunks() ) ? my_buffer->numChunks() - 1: dd;
	
	// get parms
	int Dry = (int) me->parms[2];
	int Wet = (int) me->parms[1];
	int Rgn = (int) me->parms[3];
	
	short * Old = (short *) my_buffer->getChunk(dd);
	short * In = (short *) verifyIn(ME, 0);
	short * Out = (short *) newChunk();
	int nSamps = sound_chunk_size / sizeof(short);
	int i; int tmp, tot;
	for (i=0;i<nSamps;i++)
	{
		// mix Old and In into Out, based upon Wet/Dry
		// then mix Out and In back into In, based upon Rgn/1
		tmp = In[i]; tmp *= Dry; tmp /= 256;
		tot = Old[i]; tot *= Wet; tot /= 256;
		tot += tmp;
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		Out[i] = tot;
		
		tot *= Rgn; tot /= 256;
		tot += In[i];
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		In[i] = tot;
	}
	
	me->chunkIn[0] = (char *) Out;
	my_buffer->addChunk( (char *) In);
	freeChunk ( (char *) In);
	freeChunk ( (char *) Old);
	
	
	SendOut(ME);
}

void unReverb(int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	reverbBuffer * my_buffer = (reverbBuffer *) me->Extra;
	if (my_buffer != NULL) delete my_buffer;
	me->Extra = NULL;
}

void fSplit(int ME)
{
	char * IO[4]; int i;
	for (i=0;i<4;i++) IO[i]= verifyIn(ME,i);	
	for (i=1;i<4;i++) memcpy(IO[i], IO[0], sound_chunk_size);
	SendOut(ME);	
}

struct sTone
{
	int LastOut[4];
};


void iTone(int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	me->parms[0]=128.0f;
	me->parms[1]=128.0f;
	me->parms[2]=128.0f;
	// tone control works by a low-pass filter which,
	// for seamlessness between chunks, must store
	// previous output samples
	me->Extra = new sTone;
	sTone * a = (sTone *) me->Extra;
	memset (a->LastOut, 0, 4 * sizeof(int) );
	
}

void fTone(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	short * In = (short *) verifyIn(ME,0);
	sTone * old_outs = (sTone *) me->Extra;
	int Bass = (int) me->parms[1];
	int Treble = (int) me->parms[2];
	int nSamps = sound_chunk_size / sizeof(short);
	int i, j, tmp, tot, lop, hip;	
	
	
	// X is frequency where 0 is 0 and 1 is 22050 Hz.
	// good range would be 30 to 5000 Hz
	float X = 30 + 5000 * (me->parms[0]/256.0f) * (me->parms[0] / 256.0f);
	X /= 22050.0f;
	
	// now calculate coefficients for recursive 4-pole low-pass filter
	// first as floating point numbers, then convert to fixed point for speed
	// filter uses equation y[n] = (1-X)^4 * x[n] + 4X * y[n-1] - 6X^2 * y[n-2] + 4X^3 * y[n-3] - X^4 *
	//                                                                                           y[n-4]
	
	float fCoeffs[5]; fCoeffs[0] = 1.0f-X; fCoeffs[0] *= fCoeffs[0]; fCoeffs[0] *= fCoeffs[0];
	fCoeffs[1] = 4.0f*X; fCoeffs[2] = X * X * -6.0f; fCoeffs[3] = 4.0f*X*X*X;
	fCoeffs[4] = X*X; fCoeffs[4] *= fCoeffs[4]; fCoeffs[4] *= -1.0f;
	
	// but nobody wants to use floats for DSP
	int iCoeffs[5];
	for (i=0;i<5;i++) {iCoeffs[i] = int (fCoeffs[i] * 65536.0f);}
	
	for (i=0;i<nSamps;i++)
	{
		tot = ( iCoeffs[0] * In[i] ) / 65536;
		for (j=0;j<4;j++)
		{
			tmp = iCoeffs[j+1] * old_outs->LastOut[j];
			tmp /= 65536; tot += tmp;
		}
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		for (j=3; j > 0; j--) old_outs->LastOut[j] = old_outs->LastOut[j-1];
		old_outs->LastOut[0] = tot;
		lop = tot; hip = In[i] - lop;
		tot = ( lop * Bass )/ 256 + (hip * Treble) / 16;
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		In[i] = tot;
	}
	
	
	SendOut(ME);
}


struct sLooper
{
	unsigned int oldLoopPos;
	unsigned int oldLastSync;
};


void iLooper(int ME) 
{
	soundModule *me = (soundModule *)MOD[ME];
	me->parms[2]=1023.0f;
	//me->Extra = GINT_TO_POINTER(0); // start loop at beginning
	sLooper * S = new sLooper;
	S->oldLoopPos = S->oldLastSync = 0;
	me->Extra = S;
	me->parms[3]=100.0f;
	me->parms[4]=0.0f;
	me->parms[5]=1.0f;
}

void unLooper(int ME) 
{
	soundModule *me = (soundModule *)MOD[ME];
	delete (sLooper *) me->Extra;
}

void fLooper(int ME)
{
	soundModule *me = (soundModule *)MOD[ME];
	sLooper * S = (sLooper *) me->Extra;
	unsigned int loopPos = S->oldLoopPos;
	unsigned int lastSync = S->oldLastSync;
	me->chunkIn[0]=newChunk();
	memset (me->chunkIn[0],0,sound_chunk_size);
	short * Output = (short *) me->chunkIn[0];
	short * ClkIn = (short *) verifyIn(ME,3); // watch for clock pulses!
	short * ClkOut = (short *) newChunk();
	memset( (char *) ClkOut, 0, sound_chunk_size);
	int loopNum = int(me->parms[0]);
	//////  This is the first time I've used float math in one of these.  
	//////  Hope it isn't too slow.... at least no division, that always
	//////  hurts like hell!  Probably I can get away with this.......
	float Speed = me->parms[3] / 100.0f; 
	// mutex for loop had better not be on!
	if (isLockLoop(loopNum)) {freeChunk((char*) ClkOut); SendOut(ME); return;}
	int Len = loopLength[loopNum] / sizeof(short);
	if (Len == 0) {freeChunk((char*) ClkOut); SendOut(ME); return;}
	short * Sample = (short *) loopData[int(me->parms[0])];
	if (Sample == NULL) {freeChunk((char*) ClkOut); SendOut(ME); return;}
	int i;
	unsigned int Start = Len * int(me->parms[1]); Start /= 1024;
	unsigned int End = Len * int(me->parms[2]); End /= 1024;
	if (Start == End) {freeChunk((char*) ClkOut); SendOut(ME); return;}
	if (Start > End) {i = Start; Start = End; End = i;}
	if (loopPos < Start || loopPos >= End) loopPos = Start;
	unsigned int divSync;
	divSync = (End - Start) / int(me->parms[5]);

	int nSamps = sound_chunk_size / sizeof(short);
	/// don't ask why for now, FP slop is sloshing over array ends.
	float fLoopPos = (float) loopPos;
	for (i=0;i<nSamps;i++)
	{
		Output[i] = Sample[loopPos];
		fLoopPos += Speed;
		loopPos = int(fLoopPos); 
		// handle wrap-around
		if (loopPos >= End) 
		{
		// if trigger-only off, wrap around
		// otherwise hang loopPos here at the end
			if (me->parms[4] == 0.0f)
			{
				loopPos = Start;
				fLoopPos = (float) Start;
			}
			else
			{
				loopPos = End-1;
				fLoopPos = (float) loopPos;
			}
		}
		// check for sync pulse
		if (ClkIn[i] > 1024) {loopPos = Start; fLoopPos = (float) Start;}
		// send sync pulse
		if (loopPos == Start) {ClkOut[i]=32767; lastSync = loopPos;}
		if ( (loopPos - lastSync) > divSync) { ClkOut[i]=32767; lastSync = loopPos;}
	}
	S->oldLoopPos = loopPos;
	S->oldLastSync = lastSync;
	freeChunk( (char *) ClkIn); me->chunkIn[3]=(char *) ClkOut;
	SendOut(ME);
}

void fXFader(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	short * I1 = (short *) verifyIn(ME,0);
	short * I2 = (short *) verifyIn(ME,1);
	int iVol2 = int(me->parms[0]);
	int iVol1 = 256 - iVol2;
	int tot;
	int i;
	int nSamps = sound_chunk_size / sizeof(short);
	for (i=0;i<nSamps;i++)
	{
		tot = (I1[i] * iVol1)/256 + (I2[i] * iVol2) / 256;
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		I1[i] = tot;
	}
	
	SendOut(ME);
}

struct sOctave
{
	int LastSample;
	int InverterSign;
};


void iOctave(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sOctave * S = new sOctave;
	//S->Filter.SetFreq(0.5f);
	S->LastSample = 0;
	S->InverterSign = 0;
	me->Extra = S;
}

void fOctave(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	short * In = (short *) verifyIn(ME,0);
	int nSamps = sound_chunk_size / sizeof(short);
	int i;
	sOctave * S = (sOctave *) me->Extra;
	int LastSample = S->LastSample;
	int InverterSign = S->InverterSign;
	
	for (i=0;i<nSamps;i++)
	{
		// check for zero crossing
		if (LastSample < 0 && In[i] >=0) InverterSign ^= 1;
		LastSample = In[i];
		if (InverterSign) In[i]=-In[i];
	//	In[i]=S->Filter.Go(In[i]);
	//	In[i] = (InverterSign)?8000:-8000;
	}
	//me->Extra = GINT_TO_POINTER( (InverterSign<<1) || (LastSample >> 15) );
	S->LastSample = LastSample;
	S->InverterSign = InverterSign;
	SendOut(ME);
}

void unOctave(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sOctave * S = (sOctave *)me->Extra;
	delete S;
}

struct sRec
{
	short * Data;
	int nData;
	int curPos;
};


void iRec1(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = new sRec;
	SR->Data = new short [44100];
	memset(SR->Data, 0, 44100*sizeof(short));
	SR->nData = 44100;
	SR->curPos = 0;
	me->Extra = SR;
}

void iRec0(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = new sRec;
	SR->Data = new short [441000];
	memset(SR->Data, 0, 441000*sizeof(short));
	SR->nData = 441000;
	SR->curPos = 0;
	me->Extra = SR;
}

void iRec3(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = new sRec;
	SR->Data = new short [30*44100];
	memset(SR->Data, 0, 30*44100*sizeof(short));
	SR->nData = 30*44100;
	SR->curPos = 0;
	me->Extra = SR;
}


void unRec(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = (sRec *) me->Extra;
	if (SR != NULL)
	{
		delete SR->Data;
		delete SR;
		SR = NULL;
	};
}

void fRec(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = (sRec *) me->Extra;
	int Mode = int(me->parms[0]);
	short * In = (short *) verifyIn(ME, 0);
	short * Clk = (short *) verifyIn(ME,3);
	int i, curs, tmp;
	int nSamps = sound_chunk_size / sizeof(short);
	switch (Mode)
	{
	case 0: // just play whatever we have, watch Clk
		curs = SR->curPos;
		for (i=0;i<nSamps;i++)
		{
			tmp=SR->Data[curs] + In[i];
			tmp = (tmp < -32767) ? -32767 : (tmp > 32767) ? 32767 : tmp;
			In[i] = tmp;
			curs++; if (curs >= SR->nData) curs=0;
			if (Clk[i] > 1024) curs = 0;
			if (curs == 0) Clk[i] = 32767;
		}
		SR->curPos = curs;
		break;
	case 1: // overdub over what we have, play result, watch Clk
		curs = SR->curPos;
		for (i=0;i<nSamps;i++)
		{
			tmp = In[i] + SR->Data[curs];
			tmp = (tmp < -32767) ? -32767 : (tmp > 32767) ? 32767 : tmp;
			In[i] = SR->Data[curs] = tmp;
			curs++; if (curs >= SR->nData) curs=0;
			if (Clk[i] > 1024) curs = 0;
			if (curs == 0) Clk[i] = 32767;
		}
		SR->curPos = curs;
		break;
	case 2: // replace what we have, play it, watch Clk
		curs = SR->curPos;
		for (i=0;i<nSamps;i++)
		{
			SR->Data[curs] = In[i];
			curs++; if (curs >= SR->nData) curs=0;
			if (Clk[i] > 1024) curs = 0;
			if (curs == 0) Clk[i] = 32767;
		}
		SR->curPos = curs;
		break;
	}
	
	SendOut(ME);
}

void getfromRec(int ME, short **pData, int * pLen)
{
	soundModule * me = (soundModule *)MOD[ME];
	sRec * SR = (sRec *) me->Extra;
	*pData = SR->Data;
	*pLen = SR->nData;
}
	
void iSeq(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	me->Extra = GINT_TO_POINTER(0);
}

void fSeq(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	int lastPos = GPOINTER_TO_INT(me->Extra);
	short * Clk = (short *) verifyIn(ME,3);
	int nSamps = sound_chunk_size / sizeof(short);
	int i;
	for (i=0;i<nSamps;i++)
	{
		// hit a clock pulse
		if (Clk[i] > 1024)
		{
			// toss pulse away?
			if (me->parms[lastPos] < 0.1f) Clk[i]=0;
			lastPos++; lastPos &= 7;
		}
		
	}
	me->Extra = GINT_TO_POINTER(lastPos);
	SendOut(ME);
}

void unSeq(int ME){;}

void iPan(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	int i;
	for (i=0;i<4;i++) me->parms[i]=200.0f;
	for (;i<8;i++) me->parms[i]=128.0f;
}

void fPan(int ME)
{
	soundModule * me = (soundModule *)MOD[ME];
	int nSamps = sound_chunk_size / sizeof(short);
	short * In[4]; int i;
	for (i=0;i<4;i++) In[i]= (short *) verifyIn(ME,i);
	short * Left, * Right;
	Left = (short *) newChunk(); Right = (short *) newChunk();
	int Vol[8]; float VL, VR, V;
	for (i=0;i<4;i++)
	{
		VL = me->parms[i]; VR = me->parms[i];
		V = me->parms[i+4]; VL /= 256.0f; VR /= 256.0f;
		if (V < 128.0f)
			VR *= ( V / 128.0f);
		else
			VL *= ( (256.0f - V) / 128.0f);
		Vol[i] = int (VL * 65536.0f);
		Vol[i+4] = int (VR * 65536.0f);
	}
	int tmp, tot, j;
	for (i=0;i<nSamps;i++)
	{
		tot = 0;
		for (j=0;j<4;j++)
		{
			tmp = *In[j]; tmp *= Vol[j]; tmp /= 65536;
			tot += tmp;
		}
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		Left[i] = tot;
		tot = 0;
		for (j=0;j<4;j++)
		{
			tmp = *In[j]; tmp *= Vol[j+4]; tmp /= 65536;
			tot += tmp; In[j]++;
		}
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		Right[i] = tot;
		
	}
	freeChunk(me->chunkIn[0]); freeChunk(me->chunkIn[1]);
	me->chunkIn[0]=(char *) Left; me->chunkIn[1]= (char *) Right;
	SendOut(ME);
}


struct sChor
{
	backBuf * memory;
	float ang;
};

int sinLookUp [360];
gboolean isSinLookUp = FALSE;

static void initSinLookUp()
{
	if (isSinLookUp) return;
	cout << "Initializing sin lookup table" << endl;
	int i; float arg, val;
	for (i=0;i<360;i++)
	{
		arg = ( float(i) * 3.14159265f ) / 180.0f;
		val = sin(arg);
		sinLookUp[i] = int (val * 1024);
	}
	isSinLookUp = TRUE;
}


void iChor(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sChor * S = new sChor;
	S->memory = new backBuf (44100); // 1 second memory
	me->Extra = S;
	S->ang = 0.0f;
	me->parms[0] = 50;
	me->parms[1] = 5;
	me->parms[2] = 0;
	me->parms[3] = 100;
	me->parms[4] = 200;
	initSinLookUp();
}

void fChor(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sChor * S = (sChor *) me->Extra;
	
	int nSamps = sound_chunk_size / sizeof(short);
	int i;
	int Wet = int(me->parms[3]); int Dry = int(me->parms[4]);
	int Rgn = int(me->parms[5]);
	int MaxDly = int (me->parms[0]) * 8;
	int Mode = (me->parms[2] > 0.1f) ? 1 : 0;
	float ang = S->ang;
	float angInc = me->parms[1] / 1000.0f;
	int dly = 0;
	// process the samples
	short * In = (short *) verifyIn(ME, 0);
	int tmp, tot, rgn;
	for (i=0;i<nSamps;i++)
	{
		tmp = In[i]; tmp *= Dry; tmp /= 256;
		switch (Mode)
		{
		case 0: // chorus
			dly = MaxDly * ( 1024 + sinLookUp [ (int) ang ]) ;
			dly /= 2048; ang += angInc; if (ang >= 359.0f) ang = 0.0f;
			if (dly < 0) dly = 0;
			break;
		case 1: // flange
			dly = 16 * MaxDly * ( 1024 + sinLookUp [ (int) ang ]/16);
			dly /= 2048; ang += angInc; if (ang >= 359.0f) ang = 0.0f;
			if (dly < 0) dly = 0;
			break;
		};
		
		tot = S->memory->get( (unsigned int) dly);
		tot *= Wet; tot /=256;
		tot += tmp;
		tot = (tot < -32767) ? -32767 : (tot > 32767) ? 32767 : tot;
		rgn = (S->memory->get( (unsigned int) dly) * Rgn) / 256 + In[i];
		rgn = (rgn < -32767) ? -32767 : (rgn > 32767) ? 32767 : rgn;	
		S->memory->add(rgn);
		In[i] = tot;
	}
	
	S->ang = ang;
		
	SendOut(ME);
}

void unChor(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sChor * S = (sChor *) me->Extra;
	delete S->memory; 
	delete S;
	me->Extra = NULL;
	
}

struct sComp
{
	int volAccum;
};


void iComp(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sComp * S = new sComp;
	me->Extra = S;
	me->parms[0] = 0.0f;
	me->parms[1] = 10.0f;
	me->parms[2] = 40.0f;
}

void unComp(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sComp * S = (sComp *) me->Extra;
	delete S;
}

void fComp(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sComp * S = (sComp *) me->Extra;
	int volAccum = S->volAccum;
	int nSamps = sound_chunk_size / sizeof(short);
	short * In = (short *) verifyIn(ME,0);
	int Thresh = int(me->parms[0]);
	int Comp = int(me->parms[1]);
	float CompW1 = Comp / 100.0f;
	float CompW2 = 1.0f - CompW1;
	int TimeConstant = int(me->parms[2]);
	float gateFac; float compFac;
	int i, tmp;
	for (i=0;i<nSamps;i++)
	{
		// update volAccum
		tmp = In[i]; tmp = (tmp < 0) ? -tmp : tmp;
		volAccum = (256 - TimeConstant) * volAccum + TimeConstant * tmp;
		volAccum /= 256;
		
		// handle compression
		compFac = 30000.0f / float(volAccum);
		compFac = CompW1 * compFac + CompW2;
		// handle gate
		if (Thresh <= 1.0f) gateFac = 1.0f;
		else
			gateFac = (volAccum > (Thresh * 100) ) ? 1.0f : 
					(float(volAccum) / float(Thresh * 100) );
		// process signal...
		tmp = int(float(In[i]) * compFac * gateFac);
		tmp = (tmp < -32767) ? -32767 : (tmp > 32767) ? 32767 : tmp;
		In[i] = tmp;
		 
	}
	S->volAccum = volAccum;	
	SendOut(ME);
}

struct sScratch
{
	float lastPos;
	backBuf * buf;
};


void iScratch(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sScratch * S = new sScratch;
	S->lastPos = 0.0f;
	S->buf = new backBuf(44100); // 1 second scratch buffer
	me->Extra = S;
}
 
void fScratch(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sScratch * S = (sScratch *) me->Extra;
	float newPos = me->parms[0];
	float oldPos = S->lastPos;
	int nSamps = sound_chunk_size / sizeof(short);
	short * In = (short *) verifyIn(ME,0);
	newPos *= 170.0f; oldPos *= 170.0f;
	float incPos = (newPos - oldPos)/ float(nSamps);
	int i;
	for (i=0;i<nSamps;i++)
	{
		S->buf->add(In[i]);
		In[i] = S->buf->get(int(oldPos));
		oldPos += incPos;
	}
	S->lastPos = me->parms[0];
	SendOut(ME);
}

void unScratch(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	sScratch * S = (sScratch *) me->Extra;
	delete S->buf;
	delete S;
	
}

void iMult(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	me->parms[0] = 128.0f;
	me->parms[1] = 64.0f;
}


void fMult(int ME)
{
	soundModule * me = (soundModule *) MOD[ME];
	float fBias = me->parms[0];
	float fScale = me->parms[1];
	int iBias = int(fBias);
	int iScale = int(fScale);
	int tmp;
	int nSamps = sound_chunk_size / sizeof(short);
	short * Signal = (short *) verifyIn(ME,0);
	short * Modulate = (short *) verifyIn(ME,1);
	int i;
	for (i=0;i<nSamps;i++)
	{
		tmp = Modulate[i] * iScale; tmp /= 256;
		tmp += iBias * 64;
		tmp *= Signal[i]; tmp /= 32768;
		tmp = (tmp < -32767) ? -32767 : (tmp > 32767) ? 32767 : tmp;
		Signal[i]=tmp;
	}
	
	SendOut(ME);
}

