/////////////////////////////////////////////////////////////////////////
#include <gtk/gtk.h>
#include <iostream.h>
#include "soundNode.h"
#include <pthread.h>
#include <semaphore.h>
#include "pump.h"

volatile int I_TERM_CONNECT[TERMS_I_OR_O];
volatile int O_TERM_CONNECT[TERMS_I_OR_O];
int MOD_TIER[MAX_MODULES];
int MOD_NUM_IN_TIER[MAX_MODULES];



gboolean MOD_PARMBOX[MAX_MODULES];
GtkWidget * MOD_PARMBOX_HANDLE[MAX_MODULES];

int MAX_TIER;
int MAX_NUM_IN_TIER;

int sound_buffer_size = 128;
int sound_chunk_size = 64;


volatile soundModule * MOD[MAX_MODULES];
const int BIG_TIER = 99999;

typeDesc typeList[] =
{
	{"Line In", 0x0C, "", fLineIn, NULL, NULL},
	{"Line Out", 0xC0, "", fLineOut, NULL, NULL},
	{"Passthrough", 0xFF, "!VolumeA/VolumeB/VolumeC/VolumeD", fPass, iPass, NULL},
	{"Mixer", 0xF8, "!VolumeA/VolumeB/VolumeC/VolumeD", fMixer, iMixer, NULL},
	{"Distort", 0x88, "!Drive/Sat/Level", fDistort, iDistort, NULL},
	{"Reverb", 0x88, "!Delay/Wet/Dry/Regen", fReverb, iReverb, unReverb},
	{"Splitter", 0x8F, "", fSplit, NULL, NULL},
	{"Tone" , 0x88, "!Frequency/Bass/Treble", fTone, iTone, NULL},
	{"Looper", 0x19, "#Sample/0/63/Start/0/1024/End/0/1024/Speed%/10/800/TrigOnly/0/1/Trigger Beats/1/8", 
			fLooper, iLooper, unLooper},
	{"XFader", 0xC8, "!Fade", fXFader, NULL, NULL},
	{"Octave Down", 0x88, "", fOctave, iOctave, unOctave},
	{"Record Sample (1 sec)", 0x99, "%Play/Overdub/Record", fRec, iRec1, unRec},
	{"Record Sample (10 sec)", 0x99, "%Play/Overdub/Record", fRec, iRec0, unRec},
	{"Record Sample (30 sec)", 0x99, "%Play/Overdub/Record", fRec, iRec3, unRec},
	{"Sync Sequencer", 0x11, "#1/0/1/2/0/1/3/0/1/4/0/1/5/0/1/6/0/1/7/0/1/8/0/1",
			fSeq, iSeq, unSeq},
	{"Panning Mixer", 0xFC, "!VolA/VolB/VolC/VolD/PanA/PanB/PanC/PanD", fPan, iPan, NULL},
	{"Chorus/Flange", 0x88,
	"#Depth/0/100/Speed/0/100/Mode/0/1/Wet/0/256/Dry/0/256/Regen/0/256", 
			fChor, iChor, unChor},
	{"Compressor", 0x88, "#Noise Threshold/0/256/Compression/0/100/Gate Width/1/256",
			fComp, iComp, unComp},
	{"Scratcher", 0x88, "!Move this", fScratch, iScratch, unScratch},
	{"Multiply", 0xC8, "!Bias/Scale", fMult, iMult, NULL}
};

int MAX_TYPES = 20;

volatile int MOD_EXEC_ORDER[MAX_MODULES];
volatile int MOD_EXEC_NUM;


static void recurseModuleTiers(int modNum, int tierLevel);

gint compareModExec(gconstpointer a, gconstpointer b);

void assignModuleTiers()
{
	int TIER_COUNT[MAX_MODULES];
	gboolean MOD_LONER[MAX_MODULES];
	gboolean LoneFlag;
	MAX_TIER=0; MAX_NUM_IN_TIER=0;
	int i, j;
	for (i=0;i<MAX_MODULES;i++) MOD_LONER[i]=0;
	// search for modules with no inputs - these are sources of audio //
	// and will be Tier 0 //////////////////////////////////////////////
	for (i=0;i<MAX_MODULES;i++)
	{
		LoneFlag = TRUE;
		for (j=0;j<4;j++)
			if (is_I_WIRED(IO_FROM_MOD_TERM(i,j))) LoneFlag = FALSE;	
		MOD_LONER[i]=LoneFlag;
	}
	
	// set all tier levels to bignum
	for (i=0;i<MAX_MODULES;i++) MOD_TIER[i]=BIG_TIER;
	// call recursive tier-assignment
	for (i=0;i<MAX_MODULES;i++)
		if (MOD_LONER[i]) recurseModuleTiers(i,0);
	// set remaining bignum-tier modules, which must not be connected
	// to any inputs, to zero.  assign MAX_TIER.
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD_TIER[i]==BIG_TIER) MOD_TIER[i]=0;
		if (MOD_TIER[i]>MAX_TIER) MAX_TIER=MOD_TIER[i];
	}
	// now give each module a number within its tier
	for (i=0;i<MAX_MODULES;i++) TIER_COUNT[i]=0;
	for (i=0;i<MAX_MODULES;i++)
	{
		j = MOD_TIER[i]; 
		if (j==-1) continue;
		if (MOD[i]==NULL) continue;
		MOD_NUM_IN_TIER[i]=TIER_COUNT[j]; 
		TIER_COUNT[j]++;
		if (TIER_COUNT[j]>MAX_NUM_IN_TIER)
			MAX_NUM_IN_TIER=TIER_COUNT[j];
	}
	
	// build execution order for modules.
	// put module numbers in list, then sort list by TIER_COUNT and MOD_NUM_IN_TIER
	GSList * execList = NULL;
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD[i]==NULL) continue;
		execList = g_slist_prepend(execList, GINT_TO_POINTER(i));
	}
	
	execList = g_slist_sort(execList, compareModExec);
	MOD_EXEC_NUM=0;
	GSList * p = execList;
	while (p != NULL)
	{
		MOD_EXEC_ORDER[MOD_EXEC_NUM]=GPOINTER_TO_INT(p->data);
		MOD_EXEC_NUM++; p = g_slist_next(p);
	}
	g_slist_free(execList);
	// debug - list execution order
	cout << "MOD_EXEC_NUM " << MOD_EXEC_NUM << endl;
	for (i=0;i<MOD_EXEC_NUM;i++)
	{
		cout << "Module ID " << MOD[MOD_EXEC_ORDER[i]]->UID << " executes " << i << endl;
	}
	
}

gint compareModExec(gconstpointer a, gconstpointer b)
{
	int M1 = GPOINTER_TO_INT(a);
	int M2 = GPOINTER_TO_INT(b);
	g_assert(M1 >= 0 && M1 < MAX_MODULES);
	g_assert(M2 >= 0 && M2 < MAX_MODULES);
	if (MOD_TIER[M1]<MOD_TIER[M2]) return -1; // M1 < M2
	if (MOD_TIER[M1]>MOD_TIER[M2]) return  1; // M2 > M1
	if (MOD_NUM_IN_TIER[M1]<MOD_NUM_IN_TIER[M2]) return -1; 
	return 1;
}



static void recurseModuleTiers(int modNum, int tierLevel)
{
	if (tierLevel >= MOD_TIER[modNum]) return;
	MOD_TIER[modNum] = tierLevel;
	int i; int OutTo;
	for (i=0;i<4;i++)
	{
		OutTo = O_TERM_WIRE(IO_FROM_MOD_TERM(modNum,i));
		if (OutTo != -1)
		{
			recurseModuleTiers(MOD_FROM_IO(OutTo),tierLevel+1);
		}
		
	}	
}

void validateModuleConnections()
{
	int i; int hin, her;
	for (i=0;i<TERMS_I_OR_O;i++)
	{
		hin = O_TERM_WIRE(i);
		if (hin != -1)
		{
			her = I_TERM_WIRE(hin);
			if (i != her) // mismatch!!!!
			{
				O_TERM_CONNECT[i] = -1;
				I_TERM_CONNECT[hin] = -1;
				cout << "OTERM " << i << " -> ITERM " << hin 
					<< ":" << "ITERM " << hin << " -> OTERM " << her << endl;
				cout << "Deleting mismatched wires" << endl;
			}
		}
		hin = I_TERM_WIRE(i);
		if (hin != -1)
		{
			her = O_TERM_WIRE(hin);
			if (i != her) // mismatch!!!!
			{
				I_TERM_CONNECT[i] = -1;
				O_TERM_CONNECT[hin] = -1;
				cout << "ITERM " << i << " -> OTERM " << hin 
					<< ":" << "OTERM " << hin << " -> ITERM " << her << endl;
				cout << "Deleting mismatched wires" << endl;
			}
		}
	}
	
}


pthread_mutex_t modgraph_mutex;


void GET_MODGRAPH_LOCK() 
{
	pthread_mutex_lock(&modgraph_mutex);
}
void RELEASE_MODGRAPH_LOCK()
{
	pthread_mutex_unlock(&modgraph_mutex);
}





void Init_SoundNode_Data()
{
	int i; 
	for (i=0;i<TERMS_I_OR_O;i++)
		I_TERM_CONNECT[i]=O_TERM_CONNECT[i]=-1;
	for (i=0;i<MAX_MODULES;i++) {MOD[i]=NULL; MOD_PARMBOX[i]=FALSE;}
	MOD[0] = new soundModule ; 
	MOD[0]->Type = 0;
	MOD[1] = new soundModule; 
	MOD[1]->Type = 1;
	assignModuleTiers();
	cout << "Creating mutex" << endl;
	pthread_mutex_init(&modgraph_mutex,NULL);
}

void Clean_SoundNode_Data()
{
	cout << "Destroying mutex" << endl;
	pthread_mutex_destroy(&modgraph_mutex);
}


void New_Mod(int Type)
{
	if (Type < 0 || Type >= MAX_TYPES)
	{
		cout << "Error: Type " << Type << " is out of range." << endl;
		return;
	}
	
	if (Type < 2)
	{
		cout << "Can't add or remove Line In or Line Out modules." << endl;
		return;
	}
	
	
	int i;
	for (i=0;i<MAX_MODULES;i++)
	{
		if (MOD[i]==NULL) break;
	}
	if (i == MAX_MODULES)
	{
		cout << "Error: Maximum number of " << i << "modules reached." << endl;
		return;
	}
	GET_MODGRAPH_LOCK();
	MOD[i]=new soundModule;
	MOD[i]->Type = Type;
	assignModuleTiers();
	Init_Module(i);
	RELEASE_MODGRAPH_LOCK();
	cout << "Created new module ID=" << MOD[i]->UID << " Slot=" 
			<< i << " Type=" << Type << endl;
	cout << "Name of new module: " << typeList[Type].name << endl;
	
}

const int NUM_CHUNKS_IN_BLOCK = MAX_MODULES; //hope this is efficient

GMemChunk * chunkHeap;

void Init_Chunkheap()
{
	chunkHeap = g_mem_chunk_new("chunks for passing around",
			sound_chunk_size,
			sound_chunk_size*NUM_CHUNKS_IN_BLOCK,
			G_ALLOC_AND_FREE);
}
	

void Pump_Module( int M)
{
	soundModule * sM = (soundModule *) MOD[M];
	g_assert(sM != NULL);
	void (*p_funk)(int);
	p_funk = typeList[sM->Type].pumpFunc;
	if (p_funk != NULL)
		p_funk(M);
	
}

void Clean_Module( int M)
{
	soundModule * sM = (soundModule *) MOD[M];
	g_assert(sM != NULL);
	void (*un_funk)(int);
	un_funk = typeList[sM->Type].cleanFunc;
	if (un_funk != NULL)
		un_funk(M);
	
}

void Init_Module ( int M)
{
	soundModule * sM = (soundModule *) MOD[M];
	g_assert(sM != NULL);
	void (*in_funk)(int);
	in_funk = typeList[sM->Type].initFunc;
	if (in_funk != NULL) in_funk(M);
}
