///////////////////////////////////////////////////////////////////////////
// each module has 4 input and 4 output terminals /////////////////////////
// maximum terminals for input or output is 256, or ///////////////////////
// maximum total modules is 256/4 = 64 (not 256/8!) ///////////////////////
///////////////////////////////////////////////////////////////////////////


const int TERMS_PER_MODULE  = 4;
const int TERMS_I_OR_O = 256; 
const int MAX_MODULES = TERMS_I_OR_O/TERMS_PER_MODULE;

void Init_SoundNode_Data();
void Clean_SoundNode_Data();


extern volatile int I_TERM_CONNECT[TERMS_I_OR_O];
extern volatile int O_TERM_CONNECT[TERMS_I_OR_O];

extern int MOD_TIER[MAX_MODULES];
extern int MOD_NUM_IN_TIER[MAX_MODULES];
extern volatile int MOD_EXEC_ORDER[MAX_MODULES];
extern volatile int MOD_EXEC_NUM;


extern gboolean MOD_PARMBOX[MAX_MODULES];
extern GtkWidget * MOD_PARMBOX_HANDLE[MAX_MODULES];

extern int MAX_TIER;
extern int MAX_NUM_IN_TIER;

extern int sound_buffer_size;
extern int sound_chunk_size;
extern GMemChunk * chunkHeap;

#include "io_term_inlines.h"

inline char * newChunk()
	{return (char *)g_malloc0(sound_chunk_size);}
	
inline void freeChunk(char * p)
	{g_free (p);}

	

static int sUID = 100;

class soundModule
{
public:
	int Type;
	int UID;
	float parms[16];
	soundModule()
	{
		Type = 2; /* passthrough as default */
		UID=sUID++;
		int i; for (i=0;i<16;i++) parms[i]=0.0f;
		for (i=0;i<4;i++) chunkIn[i]=(char *)NULL;
		Extra = NULL;
	};
	gpointer Extra;
	char * chunkIn[4];
};

extern volatile soundModule * MOD[MAX_MODULES];


void assignModuleTiers();

void validateModuleConnections();


struct typeDesc
{
	char name[32];
	char termMask;
	char parmDesc[512];
	void (*pumpFunc)(int);
	void (*initFunc)(int);
	void (*cleanFunc)(int);
};

extern typeDesc typeList[];
extern int MAX_TYPES;

void GET_MODGRAPH_LOCK();
void RELEASE_MODGRAPH_LOCK();

void New_Mod(int Type);

void Init_Chunkheap();

void Pump_Module(int M);

void Clean_Module(int M);

void Init_Module(int M);
